#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/ide.h>
#include <linux/version.h>

#include "lt_hotswap.h"

static char	*lths_drvtitle = "Laptop ultrabay hotswap driver";
static char	*lths_drvversion = "0.3.6";

#define DEBUG_FATAL_ERROR	0
#define DEBUG_ERROR		1
#define DEBUG_NORMAL		2
#define DEBUG_VERBOSE		3
static int 	debug=DEBUG_NORMAL;
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Laptop Hotswap Debug");
#define lths_printk(d, str...) if (debug >= d) printk( "lt_hotswap: " str)

static int	auto_eject=0;
module_param(auto_eject, int, 0);
MODULE_PARM_DESC(auto_eject, "Try to disable/eject IDE devices immediately on eject request");


/*
 *	Define the module agruments.
 */
ACPI_MODULE_NAME("acpi_lt_hotswap");

MODULE_AUTHOR("Jim Duchek <jim.duchek@gmail.com>");
MODULE_DESCRIPTION("Laptop Hotswap Driver");
MODULE_LICENSE("GPL");

static ssize_t lths_proc_write(struct file *file, const char __user * buffer, size_t count, loff_t *ppos);

static struct file_operations lths_proc_fops = {
  .write = lths_proc_write,
};

static unsigned long ide_get_or_set_dma_base (ide_hwif_t *hwif);



static void lths_battery_notify(acpi_handle handle, u32 event, void *data);
static void lths_ide_notify(acpi_handle handle, u32 event, void *data);
static void lths_ide_eject(struct lths_handler *lths_h);
static void lths_floppy_notify(acpi_handle handle, u32 event, void *data);
static void lths_dock_notify(acpi_handle handle, u32 event, void *data);
static void lths_dock_eject(struct lths_handler *lths_h);
static void lths_dock_probe(struct lths_handler *lths_h);
static void lths_install_handler(struct lths_handler *lths_h);
static void lths_remove_handler(struct lths_handler *lths_h);
static void lths_misc_eject(struct lths_handler *lths_h);


static struct lths_handler *lths_handler_head = NULL;

static struct lths_device lths_devices[] = {
      { "BAT1", lths_battery_notify, NULL, 0,
        {  "\\_SB.PCI0.ISA.EC0.BAT1",
	   "\\_SB.PCI.ISA.EC.BAT1",
	   "\\_SB.PCI0.ISA0.EC0.BAT1",
	   "\\_SB.PCI0.ISA.EC.BAT1",
	   "\\_SB.PCI0.AD4S.EC0.BAT1",
	   "\\_SB.PCI0.ICH3.EC0.BAT1",
	   "\\_SB.PCI0.LPC.EC.BAT1",
	   NULL,
        },
      },
      { "PMST", lths_ide_notify, lths_ide_eject, 0,
        {  "\\_SB.PCI0.PATA.PRIM.MAST",
           NULL,
        },
      },
      { "MSTR", lths_ide_notify, lths_ide_eject, 1,
        {  "\\_SB.PCI.IDE.SECN.MAST",
	   "\\_SB.PCI0.IDE0.IDES.IDSM",
	   "\\_SB.PCI0.IDE0.SCND.MSTR",
	   "\\_SB.PCI0.IDE0.CHN1.DRV0",
	   NULL,
        },
      },
      { "DMST", lths_ide_notify, lths_ide_eject, 0,
        {  "\\_SB.PCI0.PCI1.DOCK.IDE1.PRIM.MSTR",
           NULL,
        }, 
      }, 
      { "FLPY", lths_floppy_notify, lths_misc_eject, 0,
        {  "\\_SB.PCI0.ISA.FDC.FDD0",
           "\\_SB.PCI0.LPC.FDC.FDD0",
	   NULL,
        },
      },
      { "DOCK", lths_dock_notify, lths_dock_eject, 0,
	{  "\\_SB.GDCK",
	   "\\_SB.PCI0.DOCK",
	   "\\_SB.PCI0.PCI1.DOCK",
	   NULL,
	},
      },
      { "", NULL, NULL, 0, { NULL, }, },
};

        



static void
lths_setup_handlers(void)
{
  struct lths_handler *lh, *p_lh;
  struct lths_device *ld;
  acpi_status status;
  acpi_handle handle, temp;
  int p;
  
  p_lh = lh = NULL;
  
  ld = lths_devices;
  while(ld->handler != NULL) {
    p=0;
    while (p<LTHS_DEVICE_MAX_PATHS && ld->paths[p] != NULL) {
      status = acpi_get_handle(NULL, ld->paths[p], &handle);
      if (status) {
        lths_printk(DEBUG_VERBOSE, "'%s' not found.\n", ld->paths[p]);
        p++;
        continue;
      }
      lh = (struct lths_handler *)kzalloc(sizeof(struct lths_handler), GFP_KERNEL);
      if (lh == NULL) {
        lths_printk(DEBUG_FATAL_ERROR, "Unable to allocate memory!\n");
        return;
      }
      lh->handle = handle;
      lh->handler = ld->handler;
      lh->eject = ld->eject;
      lh->path = ld->paths[p];
      lh->present = -1;
      lh->num = ld->num;
      lh->eject_ps = 0;
      
      /* Test for all 5 power states */
      temp = NULL;
      status = acpi_get_handle(lh->handle, "_EJ0", &temp);
      if (ACPI_SUCCESS(status))
      	lh->eject_ps |= PS_EJ0;

      temp = NULL;
      status = acpi_get_handle(lh->handle, "_EJ1", &temp);
      if (ACPI_SUCCESS(status))
      	lh->eject_ps |= PS_EJ1;

      temp = NULL;
      status = acpi_get_handle(lh->handle, "_EJ2", &temp);
      if (ACPI_SUCCESS(status))
      	lh->eject_ps |= PS_EJ2;

      temp = NULL;
      status = acpi_get_handle(lh->handle, "_EJ3", &temp);
      if (ACPI_SUCCESS(status))
      	lh->eject_ps |= PS_EJ3;

      temp = NULL;
      status = acpi_get_handle(lh->handle, "_EJ4", &temp);
      if (ACPI_SUCCESS(status))
      	lh->eject_ps |= PS_EJ4;

      lths_printk(DEBUG_NORMAL, "'%s' found (%s)\n", ld->paths[p], 
      				(lh->eject_ps & PS_EJ0) ? "Hot-Swappable" :
      				((lh->eject_ps) ? "Warm-Swappable" : "Non-Swappable"));


      sprintf(lh->dev_class, "lths");
      sprintf(lh->bus_id, "%s", ld->bus_id);
      
      if (lths_handler_head == NULL) {
        lths_handler_head = lh;
      } else {
        p_lh->next = lh;
      }
      p_lh = lh;
      lths_install_handler(lh);
      break;
    }
    if (lh == NULL) {
      lths_printk(DEBUG_ERROR, "Failed to find '%s'\n", ld->bus_id);
    }
    ld++;
  }
  
}

static void
lths_remove_handler(struct lths_handler *lths_h)
{
  acpi_status status;
  struct acpi_device *device;
  
  status = acpi_remove_notify_handler(lths_h->handle,
                               ACPI_ALL_NOTIFY,
                               lths_h->handler);
  if (status) {
    lths_printk(DEBUG_ERROR, "Error %04x removing our notify handler\n", status);
  }
  if (lths_h->old_handler_type) {
    status = acpi_install_notify_handler(lths_h->handle,
                                  lths_h->old_handler_type,
                                  lths_h->old_handler,
                                  lths_h->old_context);
  }
  if (status) {
    lths_printk(DEBUG_ERROR, "Error %04x installing old notify handler\n", status);
  }

  status = acpi_bus_get_device(lths_h->handle, &device);
  if (!status) {
    strcpy(device->pnp.device_class, lths_h->old_dev_class);
    strcpy(device->pnp.bus_id, lths_h->old_bus_id);
  }
  
  lths_h->handle = 0;
  lths_h->old_handler_type = 0;
  lths_h->old_handler = 0;
  lths_h->old_context = 0;
}


static void
lths_install_handler(struct lths_handler *lths_h)
{
  struct acpi_namespace_node *node;
  union acpi_operand_object *obj, *notify_obj;
  acpi_status status;
  struct acpi_device *device;
  
  if (!lths_h->handle) {
    status = acpi_get_handle(NULL, lths_h->path, &lths_h->handle);
    if (status) {
      lths_printk(DEBUG_ERROR, "'%s' not found.\n", lths_h->path);
      return;
    }  
  }
  
  status = acpi_bus_get_device(lths_h->handle, &device);
  if (status) {
    lths_h->old_dev_class[0] = '\0';
    lths_h->old_bus_id[0] = '\0';
    lths_h->present = 0;
  } else {
    if (lths_h->present == -1) {
      acpi_bus_get_status(device);
      lths_h->present = device->status.present;
    }
    strcpy(lths_h->old_dev_class, device->pnp.device_class);
    strcpy(lths_h->old_bus_id, device->pnp.bus_id);
    strcpy(device->pnp.device_class, lths_h->dev_class);
    strcpy(device->pnp.bus_id, lths_h->bus_id);
  }
  
  node = ((struct acpi_namespace_node *)lths_h->handle);

  obj = ((union acpi_operand_object *)node->object);

  if (!obj ||
      ((ACPI_GET_DESCRIPTOR_TYPE(obj) != ACPI_DESC_TYPE_OPERAND)
       && (ACPI_GET_DESCRIPTOR_TYPE(obj) !=
           ACPI_DESC_TYPE_NAMED))
      || (ACPI_GET_OBJECT_TYPE(obj) == ACPI_TYPE_LOCAL_DATA)) {
    lths_printk(DEBUG_VERBOSE, "No object found -- installing our handler.\n");
  } else {

    if (obj->common_notify.system_notify) {
      lths_h->old_handler_type |= ACPI_SYSTEM_NOTIFY;
    } 
    if (obj->common_notify.device_notify) {
      lths_h->old_handler_type |= ACPI_DEVICE_NOTIFY;
    }
  
    if (lths_h->old_handler_type) {
      lths_printk(DEBUG_VERBOSE, "Old handler found! %08x %08x %08x\n", lths_h->old_handler_type,
                       (unsigned int)obj->common_notify.system_notify, (unsigned int)obj->common_notify.device_notify);  
    
      if (lths_h->old_handler_type & ACPI_DEVICE_NOTIFY)
        notify_obj = obj->common_notify.device_notify;
      else
        notify_obj = obj->common_notify.system_notify;
        
      lths_h->old_handler = notify_obj->notify.handler;
      lths_h->old_context = notify_obj->notify.context;
    
      status = acpi_remove_notify_handler(lths_h->handle,
                                          lths_h->old_handler_type,
                                          lths_h->old_handler);
      if (status) {
        lths_printk(DEBUG_ERROR, "Error %04x removing old notify handler\n", status);
      }
    } 
  }
  status = acpi_install_notify_handler(lths_h->handle,
                                       ACPI_ALL_NOTIFY,
				         lths_h->handler, 
				         (void *)lths_h);
  if (status) {
    lths_printk(DEBUG_ERROR, "Error %04x installing our notify handler\n", status);
  }
}



static acpi_status
lths_add_device(acpi_handle handle, struct acpi_device **device, struct lths_handler *lths_h)
{
  acpi_status status;
  acpi_handle phandle;
  struct acpi_device *pdevice;

  status = acpi_get_parent(handle, &phandle);
  if (status) {
    lths_printk(DEBUG_ERROR, "Error in acpi_get_parent\n");
    return -EINVAL;
  }

  /* Get the parent device */
  status = acpi_bus_get_device(phandle, &pdevice);
  if (status) {
    lths_printk(DEBUG_ERROR, "Error in acpi_bus_get_device\n");
    return -EINVAL;
  }

  /*
   * Now add the notified device.  This creates the acpi_device
   * and invokes .add function
   */
  lths_remove_handler(lths_h);
  status = acpi_bus_add(device, pdevice, handle, ACPI_BUS_TYPE_DEVICE);
  lths_install_handler(lths_h);
  if (status) {
    lths_printk(DEBUG_ERROR, "Error in acpi_bus_add\n");
    return -EINVAL;
  }
  return 0;
}

static void
lths_misc_eject(struct lths_handler *lths_h)
{
  acpi_status status;
  acpi_handle handle = lths_h->handle;
  struct acpi_object_list arg_list;
  union acpi_object arg;
  
  // FIXME this is dicked up.  Should do this proper.

  /* Issue the _EJ0 command */
  arg_list.count = 1;
  arg_list.pointer = &arg;
  arg.type = ACPI_TYPE_INTEGER;
  arg.integer.value = 1;
  status = acpi_evaluate_object(handle,
              "_EJ0", &arg_list, NULL);
  /* Return on _EJ0 failure */
  if (ACPI_FAILURE(status)) {
    /* Try the _EJ3 command -- autodetect somehow? */
    arg_list.count = 1;
    arg_list.pointer = &arg;
    arg.type = ACPI_TYPE_INTEGER;
    arg.integer.value = 1;
    status = acpi_evaluate_object(handle,
              "_EJ3", &arg_list, NULL);
    if (ACPI_FAILURE(status)) {
      lths_printk(DEBUG_ERROR, "_EJ0 + _EJ3 Failed!\n");
      return;
    }
  }
}

static void
lths_battery_notify(acpi_handle handle, u32 event, void *data)
{
  struct lths_handler *lths_h = (struct lths_handler *)data;
  int result;
  struct acpi_device *device;
  acpi_status status;

  result = acpi_bus_get_device(handle, &device);
  if (result) {
    switch(result) {
    case -ENODEV:
      status = lths_add_device(handle, &device, lths_h);
      if (status) {
        lths_printk(DEBUG_ERROR, "add_device_error %d\n", status);
        return; 
      }     
      break;
    default:
      lths_printk(DEBUG_ERROR, "bat1 %08x\n", result);
      return;
    }
  }
  
  acpi_bus_get_status(device);
  lths_h->present = device->status.present;


  switch(event) {
  case ACPI_NOTIFY_EJECT_REQUEST:
  case ACPI_NOTIFY_DEVICE_CHECK:
  case 0x80:
    /* Just ignore -- i don't think we have to eject it*/
    break;
  default:
    lths_printk(DEBUG_NORMAL, "Unknown Battery event %08x\n", event);
    break;
  }
  
  if (lths_h->old_handler) {
    lths_printk(DEBUG_VERBOSE, "Calling old handler, %08x %08x\n", (unsigned int)lths_h->old_handler, 
                        (unsigned int)lths_h->old_context);
    lths_h->old_handler(handle, event, lths_h->old_context);
  }
}


static void
lths_scan_ide(int num)
{
  ide_pci_device_t *d = ide_hwifs[num].cds;
  struct pci_dev *dev = ide_hwifs[num].pci_dev;
    
  ide_unregister(num);
    
  // FIXME this is just horrible.  :)
  // But perhaps it's not a horrible assumption to make on modern hw
  if (num == 0)
    ide_hwifs[num].mate = &ide_hwifs[1];
  else 
    ide_hwifs[num].mate = &ide_hwifs[0];
  ide_hwifs[num].channel = num;

  if (d->init_iops) {
    d->init_iops(&ide_hwifs[num]);
  }
  if (d->autodma == NODMA)
    goto bypass_legacy_dma;

  if(d->init_setup_dma)
    d->init_setup_dma(dev, d, &ide_hwifs[num]);
  else {
    unsigned long dma_base = ide_get_or_set_dma_base(&ide_hwifs[num]);

    if (dma_base) {
      if (d->init_dma) {
        d->init_dma(&ide_hwifs[num], dma_base);
      } else {
        ide_setup_dma(&ide_hwifs[num], dma_base, 8);
      }
    }
  }

bypass_legacy_dma:
  if (d->init_hwif)
      /* Call chipset-specific routine
       * for each enabled hwif
       */
    d->init_hwif(&ide_hwifs[num]);


  probe_hwif_init(&ide_hwifs[num]);
  
  
  
  create_proc_ide_interfaces();
}

static void
lths_remove_ide(int num)
{
  ide_unregister(num);
}

static void
lths_ide_eject(struct lths_handler *lths_h)
{
  lths_printk(DEBUG_NORMAL, "Attempting to eject\n");
  lths_remove_ide(lths_h->num);
  lths_misc_eject(lths_h);
  lths_h->present = 0;
}


static void
lths_ide_notify(acpi_handle handle, u32 event, void *data)
{
  struct lths_handler *lths_h = (struct lths_handler *)data;
  int result;
  struct acpi_device *device;
  acpi_status status;
  
  result = acpi_bus_get_device(handle, &device);
  if (result) {
    switch(result) {
    case -ENODEV:
      status = lths_add_device(handle, &device, lths_h);
      if (status) {
        lths_printk(DEBUG_ERROR, "ide add_device_error %d\n", status);
        return; 
      } 
      break;
    default:
      lths_printk(DEBUG_ERROR, "ide1 %08x\n", result);
      return;
    }
  }

  result = acpi_bus_get_status(device);
  if (result) {
    lths_printk(DEBUG_ERROR, "IDE2 %08x\n", result);
    return;
  }
  


  switch(event) {
  case ACPI_NOTIFY_EJECT_REQUEST:
    lths_printk(DEBUG_NORMAL, "Requesting IDE eject!\n");
    acpi_bus_generate_event(device, event, 0);
    if (auto_eject) { // FIXME perhaps a busy check?
      lths_h->eject(lths_h);
    }
    break;
  case ACPI_NOTIFY_DEVICE_CHECK:
    if (device->status.present && (lths_h->present <= 0)) {
      lths_scan_ide(lths_h->num);
    }
    lths_h->present = device->status.present;
    acpi_bus_generate_event(device, event, device->status.present);
    break;
  default:
    lths_printk(DEBUG_NORMAL, "Unknown IDE event %08x\n", event);
    break;
  }
  
  if (lths_h->old_handler)
    lths_h->old_handler(handle, event, lths_h->old_context);
}

static void
lths_floppy_notify(acpi_handle handle, u32 event, void *data)
{
  struct lths_handler *lths_h = (struct lths_handler *)data;
  int result;
  struct acpi_device *device;
  acpi_status status;

  result = acpi_bus_get_device(handle, &device);
  if (result) {
    switch(result) {
    case -ENODEV:
      status = lths_add_device(handle, &device, lths_h);
      if (status) {
        lths_printk(DEBUG_ERROR, "floppy add_device_error %d\n", status);
        return; 
      } 
      break;
    default:
      lths_printk(DEBUG_ERROR, "floppy %08x\n", result);
      return;
    }
  }

  result = acpi_bus_get_status(device);
  if (result) {
    lths_printk(DEBUG_ERROR, "floppy2 %08x\n", result);
    return;
  }

  switch(event) {
  case ACPI_NOTIFY_EJECT_REQUEST:
    lths_printk(DEBUG_NORMAL, "Requesting floppy eject!\n");
    acpi_bus_generate_event(device, event, 0);
    break;
  case ACPI_NOTIFY_DEVICE_CHECK:
    acpi_bus_generate_event(device, event, device->status.present);
    break;
  default:
    lths_printk(DEBUG_NORMAL, "Unknown floppy event %08x\n", event);
    break;
  }
  
  if (lths_h->old_handler)
    lths_h->old_handler(handle, event, lths_h->old_context);
}

static void
lths_dock_probe(struct lths_handler *lths_h)
{
  acpi_status status;
  acpi_handle handle = lths_h->handle;
  struct acpi_object_list arg_list;
  union acpi_object arg;

  /* Issue the _DCK command */
  arg_list.count = 1;
  arg_list.pointer = &arg;
  arg.type = ACPI_TYPE_INTEGER;
  arg.integer.value = 1;
  status = acpi_evaluate_object(handle,
              "_DCK", &arg_list, NULL);
  if (ACPI_FAILURE(status)) {
    lths_printk(DEBUG_ERROR, "Unable to send _DCK(1)\n");
    return;
  }


}

static void
lths_dock_eject(struct lths_handler *lths_h)
{
  acpi_status status;
  acpi_handle handle = lths_h->handle;
  struct acpi_object_list arg_list;
  union acpi_object arg;

  /* Issue the _DCK command */
  arg_list.count = 1;
  arg_list.pointer = &arg;
  arg.type = ACPI_TYPE_INTEGER;
  arg.integer.value = 0;
  status = acpi_evaluate_object(handle,
              "_DCK", &arg_list, NULL);
  if (ACPI_FAILURE(status)) {
    lths_printk(DEBUG_ERROR, "Unable to send _DCK(0)\n");
    return;
  }
  /* Issue the _EJ0 command */  //fixme for warmswap docks
  arg_list.count = 1;
  arg_list.pointer = &arg;
  arg.type = ACPI_TYPE_INTEGER;
  arg.integer.value = 1;
  status = acpi_evaluate_object(handle,
              "_EJ0", &arg_list, NULL);
  if (ACPI_FAILURE(status)) {
    lths_printk(DEBUG_ERROR, "Unable to send _EJ0(1)\n");
    return;
  }
}


static void
lths_dock_notify(acpi_handle handle, u32 event, void *data)
{
  struct lths_handler *lths_h = (struct lths_handler *)data;
  int result;
  struct acpi_device *device;
  acpi_status status;

  result = acpi_bus_get_device(handle, &device);
  if (result) {
    switch(result) {
    case -ENODEV:
      status = lths_add_device(handle, &device, lths_h);
      if (status) {
        lths_printk(DEBUG_ERROR, "dock add_device_error %d\n", status);
        return; 
      } 
      break;
    default:
      lths_printk(DEBUG_ERROR, "dock %08x\n", result);
      return;
    }
  }

  result = acpi_bus_get_status(device);
  if (result) {
    lths_printk(DEBUG_ERROR, "dock2 %08x\n", result);
    return;
  }

  switch(event) {
  case ACPI_NOTIFY_EJECT_REQUEST:
    lths_printk(DEBUG_NORMAL, "Requesting dock eject!\n");
    acpi_bus_generate_event(device, event, 0);
    break;
  case ACPI_NOTIFY_DEVICE_CHECK:
    lths_printk(DEBUG_NORMAL, "Dock device check\n");
    acpi_bus_generate_event(device, event, device->status.present);
    break;
  case ACPI_NOTIFY_BUS_CHECK:
    lths_printk(DEBUG_NORMAL, "Dock bus check\n");
    acpi_bus_generate_event(device, event, device->status.present);
    break;
  default:
    lths_printk(DEBUG_NORMAL, "Unknown dock event %08x\n", event);
    break;
  }
  
  if (lths_h->old_handler)
    lths_h->old_handler(handle, event, lths_h->old_context);
}


static ssize_t
lths_proc_write(struct file *file, const char __user * buffer, size_t count, loff_t *ppos)
{
  char str[16];
  char *c;
  struct lths_handler *lths_h = NULL;

  if (count > sizeof(str) - 1)
    return 0;
    
  memset(str, 0, sizeof(str));
 
  if (copy_from_user(str, buffer, count))
    return -EFAULT;
    
  c = str;
  while (*c) {
    if (*c == ' ') {
      *c='\0';
      c++;
      break;
    }
    c++;
  }   
  
  if (!strlen(str) || !strlen(c))
    return -EINVAL;
    
  lths_h = lths_handler_head;
  while(lths_h) {
    if (!strcmp(str, lths_h->bus_id))
      break;
    lths_h = lths_h->next;
  }
  if (lths_h == NULL)
    return -EINVAL;
  
  if (!strncmp(c, "eject", 5)) {
    if (lths_h->eject)
      lths_h->eject(lths_h);
  } else if (!strncmp(c, "dock", 5)) { // FIXME do I need something generic here?
    lths_dock_probe(lths_h);
  }
   
  return count;
}

static void
lths_setup_proc(void)
{
  struct proc_dir_entry *entry;
  
  entry = create_proc_entry("lths", S_IFREG | S_IRUGO | S_IWUSR, acpi_root_dir);
  if (entry) 
    entry->proc_fops = &lths_proc_fops;
  else
    lths_printk(DEBUG_ERROR, "Cannot open proc entry!\n");

}

static void
lths_remove_proc(void)
{
  remove_proc_entry("lths", acpi_root_dir);
}

static int __init lths_module_init(void)
{
  printk(KERN_INFO "%s version %s\n", lths_drvtitle, lths_drvversion);
  lths_setup_proc();

  lths_setup_handlers();
  
  ide_unregister(1);
  
  return 0;
}

static void __exit lths_module_exit(void)
{
  struct lths_handler *lths_h;

  lths_remove_proc();
  
  lths_h = lths_handler_head;
  while (lths_h != NULL) {
    if (!lths_h->handle) {
      lths_h = lths_h->next;
      continue;
    }
    lths_remove_handler(lths_h);
    lths_h = lths_h->next;
  }

  
  return;
}

module_init(lths_module_init);
module_exit(lths_module_exit);


static unsigned long ide_get_or_set_dma_base (ide_hwif_t *hwif)
{
	unsigned long	dma_base = 0;
	struct pci_dev	*dev = hwif->pci_dev;

#ifdef CONFIG_BLK_DEV_IDEDMA_FORCED
	int second_chance = 0;

second_chance_to_dma:
#endif /* CONFIG_BLK_DEV_IDEDMA_FORCED */

	if (hwif->mmio)
		return hwif->dma_base;

	if (hwif->mate && hwif->mate->dma_base) {
		dma_base = hwif->mate->dma_base - (hwif->channel ? 0 : 8);
	} else {
		dma_base = pci_resource_start(dev, 4);
		if (!dma_base) {
			printk(KERN_ERR "%s: dma_base is invalid\n",
					hwif->cds->name);
		}
	}

#ifdef CONFIG_BLK_DEV_IDEDMA_FORCED
	/* FIXME - should use pci_assign_resource surely */
	if ((!dma_base) && (!second_chance)) {
		unsigned long set_bmiba = 0;
		second_chance++;
		switch(dev->vendor) {
			case PCI_VENDOR_ID_AL:
				set_bmiba = DEFAULT_BMALIBA; break;
			case PCI_VENDOR_ID_VIA:
				set_bmiba = DEFAULT_BMCRBA; break;
			case PCI_VENDOR_ID_INTEL:
				set_bmiba = DEFAULT_BMIBA; break;
			default:
				return dma_base;
		}
		pci_write_config_dword(dev, 0x20, set_bmiba|1);
		goto second_chance_to_dma;
	}
#endif /* CONFIG_BLK_DEV_IDEDMA_FORCED */

	if (dma_base) {
		u8 simplex_stat = 0;
		dma_base += hwif->channel ? 8 : 0;

		switch(dev->device) {
			case PCI_DEVICE_ID_AL_M5219:
			case PCI_DEVICE_ID_AL_M5229:
			case PCI_DEVICE_ID_AMD_VIPER_7409:
			case PCI_DEVICE_ID_CMD_643:
			case PCI_DEVICE_ID_SERVERWORKS_CSB5IDE:
			case PCI_DEVICE_ID_REVOLUTION:
				simplex_stat = hwif->INB(dma_base + 2);
				hwif->OUTB((simplex_stat&0x60),(dma_base + 2));
				simplex_stat = hwif->INB(dma_base + 2);
				if (simplex_stat & 0x80) {
					printk(KERN_INFO "%s: simplex device: "
						"DMA forced\n",
						hwif->cds->name);
				}
				break;
			default:
				/*
				 * If the device claims "simplex" DMA,
				 * this means only one of the two interfaces
				 * can be trusted with DMA at any point in time.
				 * So we should enable DMA only on one of the
				 * two interfaces.
				 */
				simplex_stat = hwif->INB(dma_base + 2);
				if (simplex_stat & 0x80) {
					/* simplex device? */
/*
 *	At this point we haven't probed the drives so we can't make the
 *	appropriate decision. Really we should defer this problem
 *	until we tune the drive then try to grab DMA ownership if we want
 *	to be the DMA end. This has to be become dynamic to handle hot
 *	plug.
 */
					if (hwif->mate && hwif->mate->dma_base) {
						printk(KERN_INFO "%s: simplex device: "
							"DMA disabled\n",
							hwif->cds->name);
						dma_base = 0;
					}
				}
		}
	}
	return dma_base;
}
